import pytest

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), ("id", "==", 1)),  # type: ignore
    (
        pytest.lazy_fixture("and_specification"),  # type: ignore
        [("id", "==", 1), ("name", "==", "test")],
    ),
    (pytest.lazy_fixture("contains_specification"), ("field", "array-contains", "test")),  # type: ignore
    (pytest.lazy_fixture("in_specification"), ("field", "in", [1, 2, 3])),  # type: ignore
    (pytest.lazy_fixture("in_empty_specification"), None),  # type: ignore
    (pytest.lazy_fixture("less_than_specification"), ("id", "<", 1)),  # type: ignore
    (pytest.lazy_fixture("less_than_equal_specification"), ("id", "<=", 1)),  # type: ignore
    (pytest.lazy_fixture("greater_than_specification"), ("id", ">", 1)),  # type: ignore
    (pytest.lazy_fixture("greater_than_equal_specification"), ("id", ">=", 1)),  # type: ignore
    (pytest.lazy_fixture("dict_specification"), [("id", "==", 1), ("test", "==", 2)]),  # type: ignore
    (pytest.lazy_fixture("empty_specification"), None),  # type: ignore
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(specification, expected):
    from fractal_specifications.contrib.google_firestore.specifications import (
        FirestoreSpecificationBuilder,
    )

    assert FirestoreSpecificationBuilder.build(specification) == expected


def test_build_error(or_specification):
    from fractal_specifications.contrib.google_firestore.specifications import (
        FirestoreSpecificationBuilder,
        SpecificationNotMappedToFirestore,
    )

    with pytest.raises(SpecificationNotMappedToFirestore):
        FirestoreSpecificationBuilder.build(or_specification)



from typing import Any, Collection

import pytest

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), {"match": {"id.keyword": 1}}),  # type: ignore
    (
        pytest.lazy_fixture("or_specification"),  # type: ignore
        {
            "bool": {
                "should": [
                    {"match": {"id.keyword": 1}},
                    {"match": {"name.keyword": "test"}},
                ]
            }
        },
    ),
    (
        pytest.lazy_fixture("and_specification"),  # type: ignore
        {
            "bool": {
                "must": [
                    {"match": {"id.keyword": 1}},
                    {"match": {"name.keyword": "test"}},
                ]
            }
        },
    ),
    (pytest.lazy_fixture("in_specification"), {"query_string": {"default_field": "field", "query": [1, 2, 3]}}),  # type: ignore
    (pytest.lazy_fixture("less_than_specification"), {"bool": {"filter": [{"range": {"id": {"lt": 1}}}]}}),  # type: ignore
    (pytest.lazy_fixture("less_than_equal_specification"), {"bool": {"filter": [{"range": {"id": {"lte": 1}}}]}}),  # type: ignore
    (pytest.lazy_fixture("greater_than_specification"), {"bool": {"filter": [{"range": {"id": {"gt": 1}}}]}}),  # type: ignore
    (pytest.lazy_fixture("greater_than_equal_specification"), {"bool": {"filter": [{"range": {"id": {"gte": 1}}}]}}),  # type: ignore
    (pytest.lazy_fixture("empty_specification"), None),  # type: ignore
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(specification, expected):
    from fractal_specifications.contrib.elasticsearch.specifications import (
        ElasticSpecificationBuilder,
    )

    assert ElasticSpecificationBuilder.build(specification) == expected


def test_specification_not_mapped():
    from fractal_specifications.contrib.elasticsearch.specifications import (
        ElasticSpecificationBuilder,
        SpecificationNotMappedToElastic,
    )
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToElastic):
        ElasticSpecificationBuilder.build(ErrorSpecification())



from typing import Any, Collection

import pytest

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), {"id": {"$eq": 1}}),  # type: ignore
    (
        pytest.lazy_fixture("or_specification"),  # type: ignore
        {"$or": [{"id": {"$eq": 1}}, {"name": {"$eq": "test"}}]},
    ),
    (
        pytest.lazy_fixture("and_specification"),  # type: ignore
        {"$and": [{"id": {"$eq": 1}}, {"name": {"$eq": "test"}}]},
    ),
    (pytest.lazy_fixture("in_specification"), {"field": {"$in": [1, 2, 3]}}),  # type: ignore
    (pytest.lazy_fixture("less_than_specification"), {"id": {"$lt": 1}}),  # type: ignore
    (pytest.lazy_fixture("less_than_equal_specification"), {"id": {"$lte": 1}}),  # type: ignore
    (pytest.lazy_fixture("greater_than_specification"), {"id": {"$gt": 1}}),  # type: ignore
    (pytest.lazy_fixture("greater_than_equal_specification"), {"id": {"$gte": 1}}),  # type: ignore
    (pytest.lazy_fixture("regex_string_match_specification"), {"id": {"$regex": ".*abc.*"}}),  # type: ignore
    (
        pytest.lazy_fixture("dict_specification"),  # type: ignore
        {"$and": [{"id": {"$eq": 1}}, {"test": {"$eq": 2}}]},
    ),
    (pytest.lazy_fixture("empty_specification"), None),  # type: ignore
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(specification, expected):
    from fractal_specifications.contrib.mongo.specifications import (
        MongoSpecificationBuilder,
    )

    assert MongoSpecificationBuilder.build(specification) == expected


def test_specification_not_mapped():
    from fractal_specifications.contrib.mongo.specifications import (
        MongoSpecificationBuilder,
        SpecificationNotMappedToMongo,
    )
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToMongo):
        MongoSpecificationBuilder.build(ErrorSpecification())



from typing import Any, Collection

import pytest
from django.db.models import Q  # type: ignore

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), Q(id=1)),  # type: ignore
    (pytest.lazy_fixture("or_specification"), Q(id=1) | Q(name="test")),  # type: ignore
    (pytest.lazy_fixture("and_specification"), Q(id=1, name="test")),  # type: ignore
    (pytest.lazy_fixture("in_specification"), Q(field__in=[1, 2, 3])),  # type: ignore
    (pytest.lazy_fixture("less_than_specification"), Q(id__lt=1)),  # type: ignore
    (pytest.lazy_fixture("less_than_equal_specification"), Q(id__lte=1)),  # type: ignore
    (pytest.lazy_fixture("greater_than_specification"), Q(id__gt=1)),  # type: ignore
    (pytest.lazy_fixture("greater_than_equal_specification"), Q(id__gte=1)),  # type: ignore
    (pytest.lazy_fixture("regex_string_match_specification"), Q(id__regex=".*abc.*")),  # type: ignore
    (pytest.lazy_fixture("is_none_specification"), Q(field__isnull=True)),  # type: ignore
    (pytest.lazy_fixture("dict_specification"), Q(id=1, test=2)),  # type: ignore
    (pytest.lazy_fixture("empty_specification"), None),  # type: ignore
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(specification, expected):
    from fractal_specifications.contrib.django.specifications import (
        DjangoOrmSpecificationBuilder,
    )

    assert DjangoOrmSpecificationBuilder.build(specification) == expected


def test_create_q_dict():
    from fractal_specifications.contrib.django.specifications import (
        DjangoOrmSpecificationBuilder,
    )

    assert isinstance(DjangoOrmSpecificationBuilder._create_q({"id": 1}), Q)
    assert DjangoOrmSpecificationBuilder._create_q({"id": 1}).children == [("id", 1)]


def test_create_q_list():
    from fractal_specifications.contrib.django.specifications import (
        DjangoOrmSpecificationBuilder,
    )

    assert isinstance(DjangoOrmSpecificationBuilder._create_q([("id", 1)]), Q)
    assert DjangoOrmSpecificationBuilder._create_q([("id", 1)]).children == [("id", 1)]


def test_specification_not_mapped():
    from fractal_specifications.contrib.django.specifications import (
        DjangoOrmSpecificationBuilder,
        SpecificationNotMappedToDjangoOrm,
    )
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToDjangoOrm):
        DjangoOrmSpecificationBuilder.build(ErrorSpecification())



from typing import Any, Collection

import pandas as pd  # type: ignore
import pytest  # type: ignore

df = pd.DataFrame(
    {
        "id": [1, 2, 3, 4],
        "name": ["test", "test", "test", "test"],
        "field": [3, 4, 5, 6],
    }
)

dfi = df.set_index(["id", "name", "field"])

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), df[df.id == 1].to_dict()),  # type: ignore
    (
        pytest.lazy_fixture("or_specification"),  # type: ignore
        (df[(df.id == 1) | (df.name == "test")]).to_dict(),
    ),
    (
        pytest.lazy_fixture("and_specification"),  # type: ignore
        (df[(df.id == 1) & (df.name == "test")]).to_dict(),
    ),
    (pytest.lazy_fixture("in_specification"), df[df.field.isin([1, 2, 3])].to_dict()),  # type: ignore
    (pytest.lazy_fixture("less_than_specification"), df[df.id < 1].to_dict()),  # type: ignore
    (pytest.lazy_fixture("less_than_equal_specification"), df[df.id <= 1].to_dict()),  # type: ignore
    (pytest.lazy_fixture("greater_than_specification"), df[df.id > 1].to_dict()),  # type: ignore
    (pytest.lazy_fixture("greater_than_equal_specification"), df[df.id >= 1].to_dict()),  # type: ignore
    (pytest.lazy_fixture("empty_specification"), None),  # type: ignore
]

specifications_index = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), dfi[dfi.index.get_level_values("id") == 1].to_dict()),  # type: ignore
    (
        pytest.lazy_fixture("or_specification"),  # type: ignore
        (
            dfi[
                (dfi.index.get_level_values("id") == 1)
                | (dfi.index.get_level_values("name") == "test")
            ]
        ).to_dict(),
    ),
    (
        pytest.lazy_fixture("and_specification"),  # type: ignore
        (
            dfi[
                (dfi.index.get_level_values("id") == 1)
                & (dfi.index.get_level_values("name") == "test")
            ]
        ).to_dict(),
    ),
    (pytest.lazy_fixture("in_specification"), dfi[dfi.index.get_level_values("field").isin([1, 2, 3])].to_dict()),  # type: ignore
    (pytest.lazy_fixture("less_than_specification"), dfi[dfi.index.get_level_values("id") < 1].to_dict()),  # type: ignore
    (pytest.lazy_fixture("less_than_equal_specification"), dfi[dfi.index.get_level_values("id") <= 1].to_dict()),  # type: ignore
    (pytest.lazy_fixture("greater_than_specification"), dfi[dfi.index.get_level_values("id") > 1].to_dict()),  # type: ignore
    (pytest.lazy_fixture("greater_than_equal_specification"), dfi[dfi.index.get_level_values("id") >= 1].to_dict()),  # type: ignore
    (pytest.lazy_fixture("empty_specification"), None),  # type: ignore
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(specification, expected):
    from fractal_specifications.contrib.pandas.specifications import (
        PandasSpecificationBuilder,
    )

    f = PandasSpecificationBuilder.build(specification)

    assert f == expected or f(df).to_dict() == expected


def test_specification_not_mapped():
    from fractal_specifications.contrib.pandas.specifications import (
        PandasSpecificationBuilder,
        SpecificationNotMappedToPandas,
    )
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToPandas):
        PandasSpecificationBuilder.build(ErrorSpecification())


@pytest.mark.parametrize("specification, expected", specifications_index)
def test_build_with_index(specification, expected):
    from fractal_specifications.contrib.pandas.specifications import (
        PandasIndexSpecificationBuilder,
    )

    f = PandasIndexSpecificationBuilder.build(specification)

    assert f == expected or f(dfi).to_dict() == expected


def test_specification_not_mapped_with_index():
    from fractal_specifications.contrib.pandas.specifications import (
        PandasIndexSpecificationBuilder,
        SpecificationNotMappedToPandas,
    )
    from fractal_specifications.generic.specification import Specification

    class ErrorSpecification(Specification):
        def is_satisfied_by(self, obj: Any) -> bool:
            return False

        def to_collection(self) -> Collection:
            return []

        def __str__(self):
            return self.__class__.__name__

    with pytest.raises(SpecificationNotMappedToPandas):
        PandasIndexSpecificationBuilder.build(ErrorSpecification())



import pytest

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), {"id": 1}),  # type: ignore
    (pytest.lazy_fixture("or_specification"), [{"id": 1}, {"name": "test"}]),  # type: ignore
    (pytest.lazy_fixture("and_specification"), {"id": 1, "name": "test"}),  # type: ignore
    (pytest.lazy_fixture("dict_specification"), {"id": 1, "test": 2}),  # type: ignore
    (pytest.lazy_fixture("empty_specification"), None),  # type: ignore
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(specification, expected):
    from fractal_specifications.contrib.sqlalchemy.specifications import (
        SqlAlchemyOrmSpecificationBuilder,
    )

    assert SqlAlchemyOrmSpecificationBuilder.build(specification) == expected


def test_build_error(greater_than_specification):
    from fractal_specifications.contrib.sqlalchemy.specifications import (
        SpecificationNotMappedToSqlAlchemyOrm,
        SqlAlchemyOrmSpecificationBuilder,
    )

    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        SqlAlchemyOrmSpecificationBuilder.build(greater_than_specification)
